#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>
#include <stdio.h>
#include <stdlib.h>

#define _BACKGROUND 0xFFFFFF

/*
 * Type definitions
 */

typedef struct Resource_List_s {
    char *file;
    char *background;
} Resource_List_t;

/*
 * Global data
 */

#define tmpbg_width 4
#define tmpbg_height 4

static unsigned char tmpbg[] = {
   0x07, 0x0d, 0x0b, 0x0e};

static XtResource resources[] = {
    { "file", "File", XtRString, sizeof(String),
      XtOffset(Resource_List_t *, file), XtRString, NULL },

    { "background", "Background", XtRString, sizeof(String),
      XtOffset(Resource_List_t *, background), XtRString, NULL },
};

static XrmOptionDescRec options[] = {
    { "-f",  "*file",       XrmoptionSepArg, NULL },
    { "-bg", "*background", XrmoptionSepArg, NULL },
};

Atom __SWM_VROOT = None;

char *progname;
Display *display;
Window root;
int screen;
Colormap colormap;

char *version =
#ifdef VERSION
    VERSION;
#else
    "<no version>";
#endif

/*
 * Function Prototypes
 */

int getpixmapfromdata(char **xpmdata, Pixmap *icon, Pixmap *icon_mask, Pixel bg);
int getpixmapfromfile(char *file, Pixmap *icon, Pixmap *icon_mask, Pixel bg);
unsigned long NameToPixel(char * name);
void printusage(FILE *stream);
void printhelp(FILE *stream);
void printversion(FILE *);
Window getrootchild(Window parent);
void unsetvroot(Window win);
void setvroot(Window win);
void showtmpbg(Widget shell, Pixmap *background);
void handleevent(Widget w, XtPointer data, XEvent *event, Boolean *cont);
void exit_on_finish (Widget w);

/*
 * Set up a way for the window manager "close" function to allow the
 * app to exit gracefully.
 */

static void Finish(Widget w, XEvent * event, String * args, Cardinal * n_args) {
    exit(0);
}

static Atom wm_delete_window;
static XtTranslations wm_protocols_finish;
static XtActionsRec close_window_action[] = {
    {"finish", Finish}
};

void exit_on_finish(Widget w) {
    if(!XtIsShell(w)) {
        fprintf(stderr, "widget %s is not a shell", XtName(w));
    }

    if(!wm_delete_window) {
        wm_protocols_finish =
            XtParseTranslationTable("<Message>WM_PROTOCOLS: finish()");
        wm_delete_window =
            XInternAtom(XtDisplay(w), "WM_DELETE_WINDOW", TRUE);
        XtAppAddActions(XtWidgetToApplicationContext(w),
            close_window_action, XtNumber(close_window_action));
    }

    XtOverrideTranslations(w, wm_protocols_finish);
    XSetWMProtocols(XtDisplay(w), XtWindow(w), &wm_delete_window, 1);
}

/*
 * Functions
 */

static void showtmpbg(Widget shell, Pixmap *background) {
    Pixmap bitmap;
    GC gc;
    XGCValues gc_init;

    /* Read bitmap */
    bitmap = XCreateBitmapFromData(
        display, root, (char *)tmpbg, tmpbg_width, tmpbg_height);

    /* Create graphics context */
    gc_init.foreground = BlackPixel(display, screen);
    gc_init.background = WhitePixel(display, screen);
    gc = XCreateGC(display, root, GCForeground|GCBackground, &gc_init);

    /* Create pixmap */
    *background = XCreatePixmap(display, root, tmpbg_width, 
        tmpbg_height, (unsigned int) DefaultDepth(display, screen));

    /* Copy bitmap onto pixmap */
    XCopyPlane(display, bitmap, *background, gc, 0,0, tmpbg_width,
           tmpbg_height, 0,0, (unsigned long) 1);

    /* Set the background pixmap */
    XtVaSetValues(shell, XtNbackgroundPixmap, *background, NULL);

    /* Free bitmap */
    XFreePixmap(display, bitmap);
}

int getpixmapfromdata(char **xpmdata, Pixmap *icon, Pixmap *icon_mask, Pixel bg) {
    XpmAttributes attr;
    int result;
    XpmColorSymbol symbol;

    // Map "none" to background color
    symbol.name = 0;
    symbol.value = "none";
    symbol.pixel = bg;
    attr.colorsymbols = &symbol;
    attr.numsymbols = 1;

    attr.closeness = 65535; /* accuracy isn't crucial */
    attr.colormap = colormap;
    attr.depth = DefaultDepth(display, screen);

    attr.valuemask = XpmCloseness | XpmReturnPixels | XpmColormap
        | XpmDepth | XpmColorSymbols;

    result = XpmCreatePixmapFromData(
        display, root, xpmdata, icon, icon_mask, &attr);

    if(result != XpmSuccess) {
        fprintf(stderr, "%s: %s\n", progname, XpmGetErrorString(result));
    }

    return result;
}

/*
 * Creates a Pixmap from an XPM image file.  Prints a message to
 * STDERR if an error occurs.
 *
 * @param       *file
 *              The file to read.
 *
 * @param       *icon
 *              alloc'd memory for a Pixmap image.
 *
 * @param       *icon_mask
 *              alloc'd memory for a Pixmap image mask.
 *
 * @return      XpmOpenFailed, XpmFileInvalid, XpmNoMemory, or
 *              XpmSuccess
 */
int getpixmapfromfile(char *file, Pixmap *icon, Pixmap *icon_mask, Pixel bg) {
    char **xpmdata;
    int result = XpmReadFileToData(file, &xpmdata);

    if(result == XpmSuccess) {
        result = getpixmapfromdata(xpmdata, icon, icon_mask, bg);

        /* Free image data allocated by XpmReadFileToData */
        XpmFree(xpmdata);
    }

    switch(result) {
        case XpmOpenFailed:
            fprintf(stderr, "%s: %s: cannot open\n", progname, file);
        break;

        case XpmFileInvalid:
            fprintf(stderr, "%s: %s: invalid image file\n",
                progname, file);
        break;

        case XpmNoMemory:
            fprintf(stderr, "%s: %s: not enough memory\n",
                progname, file);
        break;

        case XpmSuccess:
        break;

        default:
            fprintf(stderr,
                "%s: %s: could not create image from file\n",
                progname, file);
        break;
    }

    return result;
}

/*
 * Convert the given color name to a XColor using the default
 * color map.
 */
Pixel NameToPixel(char * name) {
    Pixel pixel = _BACKGROUND;

    /* Was a color specified? */
    if(!name || !*name) {
        fprintf(stderr, "%s: no color specified\n", progname);
    } else {

        XColor color;
        Colormap map = DefaultColormap(display, screen);

        /* Is the color known? */
        if(!XParseColor(display, map, name, &color)) {
            fprintf(stderr, "%s: unknown color: \"%s\"\n", progname, name);
        } else {

            /* Allocate the color */
            if(!XAllocColor(display, map, &color)) {
                fprintf(stderr, "%s: unable to allocate color for \"%s\"\n", progname, name);
            } else

            pixel = color.pixel;
        }
    }

    return pixel;
}

/*
 * Print the program usage.
 */
void printusage(FILE *stream) {
    fprintf(stream, "Usage: %s [-display <display>] [-w] -f <file> [-bg <color>]\n", progname);
    fprintf(stream, "       %s [-display <display>] [-w] -bg <color>\n", progname);
    fprintf(stream, "       %s --help\n", progname);
    fprintf(stream, "       %s --version\n", progname);
}

/*
 * Print the program help.
 */
void printhelp(FILE *stream) {
    printusage(stream);

    fprintf(stream, "\nOptions:\n\n");
    fprintf(stream, "-f <file>           Specify the image file to display.  Use\n");
    fprintf(stream, "                    \"-\" to read from STDIN.\n\n");
    fprintf(stream, "-bg <color>         Set the background color of the window.\n\n");
    fprintf(stream, "-w                  Print the X window ID of the %s window.\n", progname);
    fprintf(stream, "                    Useful when using another program to draw\n");
    fprintf(stream, "                    in the %s window.\n\n", progname);
    fprintf(stream, "-display <display>  Specify the X server to contact.\n\n");
    fprintf(stream, "--help              Print this message and exit.\n\n");
    fprintf(stream, "--version           Print version information and exit.\n");
}

/*
 * Print the program version.
 */
void printversion(FILE *stream) {
    fprintf(stream, "%s %s\n", progname, version);
}

Window getrootchild(Window parent) {
    Window window, rootReturn, *children;
    unsigned int n;

    do {
        window = parent;
        XQueryTree(display, window, &rootReturn, &parent, &children, &n);
        XFree(children);
    } while(parent && parent != root);

    return window;
}

void unsetvroot(Window win) {
    XDeleteProperty(display, getrootchild(win), __SWM_VROOT);
}

void setvroot(Window win) {
    XChangeProperty(display, getrootchild(win), __SWM_VROOT,
        XA_WINDOW, 32, PropModeReplace, (unsigned char *)&win, 1);
}

Window getvroot() {
    Window rootReturn, parentReturn, *children;
    unsigned int numChildren;
    Window vroot = root;
    int i;

    XQueryTree(display, root, &rootReturn, &parentReturn, &children, &numChildren);

    for (i = 0; i < numChildren; i++) {
        Atom actual_type;
        int actual_format;
        long nitems, bytesafter;
        Window *newRoot = NULL;

        if (XGetWindowProperty(display, children[i], __SWM_VROOT, 0, 1,
            False, XA_WINDOW, &actual_type, &actual_format, &nitems,
            &bytesafter, (unsigned char **) &newRoot) == Success && newRoot)
        {
            vroot = *newRoot;
            break;
        }
    }

    XFree(children);
    return vroot;
}

/**
 * Handles the events passed to this application.  If event is a
 * ButtonPress event, send it to the (real) root window.  If event is
 * a VisibilityNotify event, set a property identifying this window as
 * the virtual root.
 *
 * Some window managers, like CDE, send a UnmapNotify event when
 * this window is no longer visible due to switching desktops.  Unset
 * the virtual root property when this event is received.
 *
 * Other window managers, like FVWM, set the "WM_STATE" property to be
 * IconicState when this window is no longer visible due to switching
 * desktops.  Unset the virtual root property when a matching
 * PropertyNotify event is received.
 */
void handleevent(Widget w, XtPointer data, XEvent *event, Boolean *cont) {
    Window window = *((Window *)data);

    switch(event->type) {
        case ButtonPress:
            XUngrabPointer(display, CurrentTime);
            event->xany.window = root;
            XSendEvent(display, root, False, ButtonPressMask, event);
            XFlush(display);
        break;

        case VisibilityNotify:
            setvroot(window);
        break;

        case UnmapNotify:
            unsetvroot(window);
        break;

        case PropertyNotify: {
            XPropertyEvent *pevent = (XPropertyEvent*)event;

            if (pevent->state == PropertyNewValue) {

                if(strcmp("WM_STATE", XGetAtomName(display, pevent->atom)) == 0) {
                    Atom actual_type;
                    int actual_format;
                    unsigned long nitems, bytes_after;
                    unsigned char *data;
                    long *state;

                    XGetWindowProperty(pevent->display,
                        pevent->window, pevent->atom, 0L, 1L, False,
                        AnyPropertyType, &actual_type, &actual_format,
                        &nitems, &bytes_after, &data);

                    state = (long *)data;
                    switch(*state) {
                        case IconicState:
                            unsetvroot(window);
                        break;
                    }
                }
            }
        }
        break;
    }
}

/*
 * Main
 */

int main(int argc, char **argv) {
    Widget shell;
    Window window;
    XtAppContext app;
    Resource_List_t list;
    Pixmap background = 0, mask = 0;
    Boolean printwin = False;
    int i;

    /* Set program name */
    progname = argv[0];
    for(i = strlen(progname) - 1; i > 0; i--) {
        if(progname[i] == '/') {
            progname += i+1;
            break;
        }
    }

    /* Initialize Xt */
    shell = XtAppInitialize(
        &app, "XVirtualRoot", options, XtNumber(options),
        &argc, argv, NULL, NULL, 0 );

    /* Check arguments */
    while(--argc) {
        int error = 0;

        if(!strcmp(argv[argc], "-w")) {
            printwin = True;
        } else

        if(!strcmp(argv[argc], "--help")) {
            printhelp(stdout);
            exit(0);
        } else

        if(!strcmp(argv[argc], "--version")) {
            printversion(stdout);
            exit(0);
        } else

        {
            printusage(stderr);
            exit(1);
        }
    }

    display = XtDisplay(shell);
    screen = DefaultScreen(display);
    root = RootWindow(display, screen);
    colormap = DefaultColormap(display, DefaultScreen(display));

    /* Set size and position */
    XtVaSetValues(shell,
        XtNwidth, DisplayWidth(display, screen),
        XtNheight, DisplayHeight(display, screen),
        XtNcolormap, colormap, NULL);

    /* Set the virtual root atom */
    __SWM_VROOT = XInternAtom(display, "__SWM_VROOT", False);

    /* Populate resource list */
    XtGetApplicationResources(
        shell, &list, resources, XtNumber(resources), 0, 0);

    /* Was a background image specified? */
    if(list.file) {
        showtmpbg(shell, &background);
    }

    /* Set up event handling */
    XtAddEventHandler(shell,
        ButtonPressMask | StructureNotifyMask |
        VisibilityChangeMask | PropertyChangeMask,
        False, (XtEventHandler)handleevent, (XtPointer)&window);

    /* Map the window */
    XtRealizeWidget(shell);
    window = XtWindow(shell);

    exit_on_finish(shell);

    /* Was a background image specified? */
    if(list.file) {
        /* Free tmp background pixmap */
        XFreePixmap(display, background);

        /* Was "-" specified as the file name? "" */
        if(strcmp(list.file, "-") == 0) {
            list.file = NULL;
        }

        if(getpixmapfromfile(list.file, &background, &mask,
            list.background ? NameToPixel(list.background) :
            _BACKGROUND) != XpmSuccess)
        {
            exit(1);
        }

        /* Set the background pixmap */
        XtVaSetValues(shell, XtNbackgroundPixmap, background, NULL);

        /* Free the pixmaps on the server */
        XFreePixmap(display, background);
        if(mask != 0) XFreePixmap(display, mask);
    }

    /* Print the window to STDOUT if requested */
    if(printwin) {
        printf("%d\n", window);
        fflush(stdout);
    }

    /* Enter event loop */
    XtAppMainLoop(app);

    return 0;
}
